class OpType:
    ANNOTATION_QUERY = "annotation_query"
    JP_ANNOTATION_QUERY = "jp_annotation_query"
    FILE_QUERY = "file_query"
    PATH_QUERY = "path_query"
    TAG_QUERY = "tag_query"
    NO_TAG_QUERY = "no_tag_query"
    LIMIT = "limit"
    SAMPLE = "sample"
    PIPE = "pipe"
